from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Union, Dict, List

from zeep.helpers import serialize_object as _so


class Invoice4UError(Exception):
    pass


@dataclass
class GeneralCustomer:
    Name: Optional[str] = None


@dataclass
class Customer:
    Name: Optional[str] = None
    ID: Optional[int] = None
    Email: Optional[str] = None
    Cell: Optional[str] = None
    Phone: Optional[str] = None
    Fax: Optional[str] = None
    Address: Optional[str] = None
    City: Optional[str] = None
    Zip: Optional[str] = None
    UniqueID: Optional[str] = None  # ת.ז או ח.פ
    PayTerms: int = -1  # מיידי
    Active: bool = True
    ExtNumber: Optional[int] = None  # מספר מזהה במערכת הפנימית
    Errors: None = None

    def __post_init__(self):
        if self.ExtNumber is None:
            self.ExtNumber = self.ID


@dataclass
class Payment:
    Amount: float
    PaymentType: int
    Date: datetime = datetime.now()


@dataclass
class CashPayment(Payment):
    PaymentType: int = 4


@dataclass
class BankTransferPayment(Payment):
    PaymentType: int = 3
    AccountNumber: str = ''
    BankName: str = ''
    BranchName: str = ''


@dataclass
class CheckPayment(BankTransferPayment):
    PaymentType: int = 2
    PaymentNumber: str = ''  # מספר המחאה


@dataclass
class CreditPayment(Payment):
    PaymentType: int = 1
    ExpirationDate: str = ''  # תוקף כרטיס
    NumberOfPayments: int = 1  # מספר תשלומים
    PayerID: str = ''  # מספר ת.ז בעל הכרטיס
    PaymentNumber: str = ''  # 4 ספרות אחרונות כרטיס אשראי


@dataclass
class DocumentItem:
    Price: float
    Name: str
    Quantity: float = 1.0
    TaxPercentage: float = 17.0

    def __post_init__(self):
        self.Price = self.Price / (1 + self.TaxPercentage / 100)


@dataclass
class Document:
    Items: Union[Dict[str, List[DocumentItem]], List[DocumentItem]]  # {'DocumentItem': []}
    Payments: Union[Dict[str, List[Payment]], List[Payment]]  # {'Payment': []}
    DocumentType: int
    Subject: Optional[str] = None
    # Total: float
    BranchID: Optional[int] = None
    ClientID: Optional[int] = None
    Currency: Optional[str] = None
    GeneralCustomer: Optional[GeneralCustomer] = None
    Is2SignDoc: Optional[bool] = None
    IssueDate: datetime = datetime.now()
    Language: int = 1
    MailsAttached: Optional[str] = None  # Comma separated list of emails
    OrganizationID: Optional[int] = None
    # SmsMessages: list
    UniqueID: Optional[str] = None
    DocumentNumber: Optional[str] = None
    Errors: None = None
    Total: Optional[float] = None
    NewCustomer: Optional[Customer] = None

    def __post_init__(self):
        if isinstance(self.Items, list):
            self.Items = {'DocumentItem': self.Items}
        if isinstance(self.Payments, list):
            self.Payments = {'Payment': self.Payments}


@dataclass
class InvoiceReceipt(Document):
    DocumentType: int = 3

    def __post_init__(self):
        super().__post_init__()
        all_items = float(sum(i.Price for i in self.Items['DocumentItem']))
        self.Total = all_items


@dataclass
class InvoiceOrder(Document):
    DocumentType: int = 6
    Payments: None = None
    Is2SignDoc: bool = True
    GeneralCustomer: None = None


@dataclass
class InvoiceQuote(InvoiceOrder):
    DocumentType: int = 7


def serialize(item, i4u_instance):
    """ Helper function to turn wsdl service objects to defined dataclasses """
    obj = _so(item)
    for key, value in obj.items():
        setattr(i4u_instance, key, value)
    return i4u_instance



# coding: windows-1255
import logging
from dataclasses import asdict
from functools import lru_cache
from pathlib import Path

import requests
import zeep

from .classes import Invoice4UError, Document, Customer, serialize

logging.getLogger('zeep').setLevel(logging.ERROR)


class Invoice4U(zeep.CachingClient):
    """ Invoice4U documentation: https://invoice4uapi.docs.apiary.io/ """
    def __init__(self, user: str, password: str, timeout=20):
        super().__init__('https://api.invoice4u.co.il/Services/ApiService.svc?singleWsdl')
        self._user = user
        self._password = password
        self.transport.operation_timeout = timeout
        self.transport.load_timeout = timeout

    @property
    @lru_cache(maxsize=1)
    def token(self):
        return self.service.VerifyLogin(self._user, self._password)

    def get_all_customers(self):
        response = self.service.GetCustomersByOrgId(self.token)
        return [serialize(c, Customer()) for c in response.Response.Customer]

    def get_customer(self, cus: Customer) -> Customer:
        if cus.ExtNumber:
            return serialize(self.service.GetCustomerByExternalNumber(cus.ExtNumber, self.token), Customer())
        if cus.Name:
            return serialize(self.service.GetCustomerByName(cus.Name, self.token), Customer())
        raise Invoice4UError('Not enough details to find customer')

    def get_document(self, doc: Document) -> Document:
        if doc.DocumentNumber and doc.DocumentType:
            zeep_doc = self.service.GetDocumentByNumber(doc.DocumentNumber, doc.DocumentType, self.token)
            i4u_doc = Document(zeep_doc.Items, zeep_doc.Payments, zeep_doc.DocumentType)
            return serialize(zeep_doc, i4u_doc)
        raise Invoice4UError('Not enough details to find document')

    @staticmethod
    def document_url(doc: Document) -> str:
        return f'https://newview.invoice4u.co.il/Views/PDF.aspx?docid={doc.UniqueID}'

    def download_document(self, doc: Document, destination: Path = Path.home() / 'Downloads') -> Path:
        url = self.document_url(doc)
        response = requests.get(url)
        if destination.is_dir():
            destination = destination / response.headers['Content-Disposition'][21:]
        destination.write_bytes(response.content)
        return destination

    def create_document(self, doc: Document) -> Document:
        if hasattr(doc, 'NewCustomer') and doc.NewCustomer is not None:
            new_customer = self.create_customer(doc.NewCustomer)
            delattr(doc, 'NewCustomer')
            doc.ClientID = new_customer.ID
        new_doc = self.service.CreateDocument(asdict(doc), self.token)
        if new_doc.Errors:
            raise Invoice4UError(new_doc.Errors)
        return serialize(new_doc, doc)

    def create_customer(self, cus: Customer) -> Customer:
        new_cus = self.service.CreateCustomer(asdict(cus), self.token)
        if new_cus.Errors:
            raise Invoice4UError(new_cus.Errors)
        return serialize(new_cus, cus)

    def update_customer(self, cus: Customer) -> Customer:
        updated = self.service.UpdateCustomer(asdict(cus), self.token)
        if updated.Errors:
            raise Invoice4UError(updated.Errors)
        return serialize(updated, cus)



from pathlib import Path

from setuptools import setup

long_description = Path(__file__).with_name('README.md').read_text(encoding='utf-8')

requirements = Path(__file__).with_name('requirements.txt').read_text().split()

setup(
    name='i4u',
    version='0.0.6',
    packages=['i4u'],
    url='https://github.com/ofersadan85/i4u',
    license='MIT License',
    author='Ofer Sadan',
    author_email='ofersadan85@gmail.com',
    description='Python package to interact with Invoice4U API',
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
    install_requires=requirements + ["dataclasses;python_version<'3.7'"],
)
